import itertools
from typing import List, Union

import numpy as np


def fstack(arrays:List[Union[List,np.ndarray]])->np.ndarray:
    """stack list/ndarray of different size into ndarray with zero padding

    :param arrays: [description]
    :type arrays: List[Union[List,np.ndarray]]
    :return: stack array
    :rtype: np.ndarray
    """
    def resize(row, size):
        new = np.array(row)
        new.resize(size)
        return new

    # find longest row length
    row_length = max(arrays, key=len).__len__()
    mat = np.array( [resize(row, row_length) for row in arrays] )

    return mat

def fstack_pad(arrays:List[Union[List,np.ndarray]],fillvalule:float=0)->np.ndarray:
    """stack list/ndarray of different size into ndarray with zero padding

    :param arrays: 
    :type arrays: List[Union[List,np.ndarray]]  
    :param fillvalule: padding value, defaults to 0
    :type fillvalule: float, optional
    :return: stack array
    :rtype: np.ndarray
    """
    return np.column_stack((itertools.zip_longest(*arrays, fillvalue=fillvalule)))



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : meta.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 01/24/2018
#
# This file is part of Jacinle.
# Distributed under terms of the MIT license.

import six
import functools
import numpy as np

import torch
import torch.distributed as dist
from jammy.utils.meta import stmap

SKIP_TYPES = six.string_types

__all__ = [
    "as_tensor",
    "as_numpy",
    "as_float",
    "as_cuda",
    "as_cpu",
    "as_detached",
    "is_master",
]


def _as_tensor(o):
    from torch.autograd import Variable

    if isinstance(o, SKIP_TYPES):
        return o
    if isinstance(o, Variable):
        return o
    if torch.is_tensor(o):
        return o
    return torch.from_numpy(np.array(o))


def as_tensor(obj):
    return stmap(_as_tensor, obj)


def _as_numpy(o):
    from torch.autograd import Variable

    if isinstance(o, SKIP_TYPES):
        return o
    if isinstance(o, Variable):
        o = o
    if torch.is_tensor(o):
        return o.cpu().numpy()
    return np.array(o)


def as_numpy(obj):
    return stmap(_as_numpy, obj)


def _as_float(o):
    if isinstance(o, SKIP_TYPES):
        return o
    if torch.is_tensor(o):
        return o.item()
    arr = as_numpy(o)
    assert arr.size == 1
    return float(arr)


def as_float(obj):
    return stmap(_as_float, obj)


def _as_cpu(o):
    from torch.autograd import Variable

    if isinstance(o, Variable) or torch.is_tensor(o):
        return o.cpu()
    return o


def as_cpu(obj):
    return stmap(_as_cpu, obj)


def _as_cuda(o):
    from torch.autograd import Variable

    if isinstance(o, Variable) or torch.is_tensor(o):
        return o.cuda()
    return o


def as_cuda(obj):
    return stmap(_as_cuda, obj)


def _as_detached(o, clone=False):
    if torch.is_tensor(o):
        if clone:
            return o.clone().detach()
        return o.detach()
    return o


def as_detached(obj, clone=False):
    return stmap(functools.partial(_as_detached, clone=clone), obj)


def is_master():
    return not dist.is_initialized() or dist.get_rank() == 0



import socket

__all__ = ["find_free_port"]

def find_free_port():
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]
    finally:
        s.close()


import functools

import matplotlib.pyplot as plt
import wandb

from .wandb_utils import Wandb

__all__ = ["wandb_plt", "wandb_img"]


def wandb_plt(func):
    @functools.wraps(func)
    def wandb_record(*args, **kwargs):
        fig, caption = func(*args, **kwargs)
        msg = None
        if Wandb.IS_ACTIVE:
            title = (
                caption
                if fig._suptitle is None  # pylint: disable= protected-access
                else fig._suptitle.get_text()  # pylint: disable= protected-access
            )
            msg = {title: wandb.Image(fig, caption=caption)}
            Wandb.log(msg)
        plt.close(fig)
        return msg

    return wandb_record


def wandb_img(title, fpath, caption=None):
    if Wandb.IS_ACTIVE:
        msg = {title: wandb.Image(fpath, caption=caption)}
        wandb.log(msg)



import torch
import torch.nn as nn

__all__ = ["MLP"]


class MLP(nn.Module):
    def __init__(
        self, input_size, n_hidden, hidden_size, output_size, is_batchnorm=False
    ):
        super().__init__()
        layers = []
        for _ in range(n_hidden):
            layers.append(nn.Linear(input_size, hidden_size))
            if is_batchnorm:
                layers.append(nn.BatchNorm1d(hidden_size))
            layers.append(nn.ReLU())
            input_size = hidden_size
        layers.append(nn.Linear(hidden_size, output_size))
        self.layers = nn.Sequential(*layers)

    def forward(self, x):
        return self.layers(x)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jammy',
 'jammy.cli',
 'jammy.collections',
 'jammy.comm',
 'jammy.event',
 'jammy.image',
 'jammy.io',
 'jammy.logging',
 'jammy.random',
 'jammy.storage.kv',
 'jammy.utils',
 'jamnp',
 'jamtorch',
 'jamtorch.cuda',
 'jamtorch.data',
 'jamtorch.ddp',
 'jamtorch.distributions',
 'jamtorch.io',
 'jamtorch.logging',
 'jamtorch.nn',
 'jamtorch.prototype',
 'jamtorch.trainer',
 'jamtorch.utils',
 'jamviz',
 'jamviz.plt',
 'jamweb',
 'jamweb.session',
 'jamweb.web']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.17,<4.0.0',
 'attrs>=20.3.0,<21.0.0',
 'einops>=0.3.0,<0.4.0',
 'gpustat>=0.6.0,<0.7.0',
 'h5py>=3.4.0,<4.0.0',
 'hydra-core>=1.1.0,<2.0.0',
 'ipdb>=0.13.8,<0.14.0',
 'lmdb>=1.2.1,<2.0.0',
 'loguru>=0.5.3,<0.6.0',
 'matplotlib>=3.3.0,<4.0.0',
 'numpy>=1.20.3,<2.0.0',
 'python-memcached>=1.59,<2.0',
 'scipy>=1.6.3,<2.0.0',
 'torch>=1.9.0,<2.0.0',
 'torchvision>=0.10.0,<0.11.0',
 'tornado>=6.1,<7.0',
 'tqdm>=4.11.0,<5.0.0',
 'wandb>=0.10.31,<0.11.0']

setup_kwargs = {
    'name': 'jammy',
    'version': '0.0.2',
    'description': 'Personal ToolBox',
    'long_description': '# Jammy (Jam)\n\n**Jammy** is a personal toolkit.\n\n\n### Etymology\n* The naming is inspired from [Jyutping](https://en.wikipedia.org/wiki/Jyutping) of [Qin](https://en.wiktionary.org/wiki/%E6%AC%BD).\n\n## MICS\n\n* The package and framework are inspired from [Jacinle](https://github.com/vacancy/Jacinle) by [vacancy](https://github.com/vacancy), from which I learn and take utility functions shamelessly.\n',
    'author': 'Qin',
    'author_email': 'qsh.zh27@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)



# Consider thread safe and singleton pattern

__all__ = ["BufferCnt", "CBCnt", "ExitCnt"]


class BufferCnt:
    def __init__(self, thres=float("inf")):
        self._cnt = 0
        self.thres = thres

    def __call__(self, expre, thres=None):
        if expre is True:
            self._cnt += 1
        else:
            self._cnt = 0

        if thres is None:
            thres = self.thres

        if self._cnt > self.thres:
            self.reset()
            return True

        return False

    @property
    def cnt(self):
        return self._cnt

    def reset(self):
        self._cnt = 0


class CBCnt(BufferCnt):
    def __init__(self, cb_fn=None, thres=float("inf")):
        super().__init__(thres)
        self.call_back = cb_fn

    def __call__(self, expre, thres=None, *args, **kwargs):
        is_thres = super().__call__(expre, thres)
        if is_thres and self.call_back is not None:
            self.call_back(*args, **kwargs)


class ExitCnt(CBCnt):
    def __init__(self, msg="Something wrong", thres=float("inf")):
        def _exit_fn():
            print(msg)
            import sys

            sys.exit()

        super().__init__(_exit_fn, thres)
